/*
  Definition for RAMC to byte code compiler.
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : Definition for RAMC to byte code compiler.
// ----------------------------------------------------------------------------




#include "compram.h"
#include "bcode.h"
#include "xstring.h"
#include "xstrre.h"
#include <string.h>




// Plain text rendering table -------------------------------------------------
#define T_INVALID_CODE          "Invalid RAM code"
#define T_INCORRECT_SYNTAX      "Syntax not correct"




static char *(PlainText[]) =
  { T_INVALID_CODE, T_INCORRECT_SYNTAX };




// xRAMCompilerException ------------------------------------------------------
xRAMCompilerException::xRAMCompilerException(TErrorCode error, TIndex line)
: xBaseException(error, NULL, NULL, 0, "RCC")
{
  HasInfo = TRUE;
  try {
    xString temp = "In line ";
    temp += xString::Unsigned2Dec(line + 1);
    strcpy(Info, temp);
  }
  catch (...) { }
}




char *xRAMCompilerException::GetText() const
{
  return PlainText[Error];
}




// RAM compiler : -------------------------------------------------------------
void CompileRAM2BC(xRAMProgram &source, xBCProgram &destination)
{
  xString       cmd;
  xByteCode     bcode;

  xRegexString  regex_mov1("R([0-9]+)<\\-R([0-9]+)");
  xRegexString  regex_mov2("R([0-9]+)<\\-RR([0-9]+)");
  xRegexString  regex_mov3("RR([0-9]+)<\\-R([0-9]+)");
  xRegexString  regex_mov4("R([0-9]+)<\\-([0-9]+)");
  xRegexString  regex_add("R([0-9]+)<\\-R([0-9]+)\\+R([0-9]+)");
  xRegexString  regex_sub("R([0-9]+)<\\-R([0-9]+)\\-R([0-9]+)");
  xRegexString  regex_goto("GOTO([0-9]+)");
  xRegexString  regex_if1("IFR([0-9]+)=0GOTO([0-9]+)");
  xRegexString  regex_if2("IFR([0-9]+)>0GOTO([0-9]+)");

  regex_mov1.Parse();
  regex_mov2.Parse();
  regex_mov3.Parse();
  regex_mov4.Parse();
  regex_add.Parse();
  regex_sub.Parse();
  regex_goto.Parse();
  regex_if1.Parse();
  regex_if2.Parse();

  for (TIndex line=0; line<~source; line++) {
    cmd = source[line];
    cmd.DeleteCharThis(' ');
    cmd.UpperThis();

    bcode.Param1 = "";
    bcode.Param2 = "";
    bcode.Param3 = "";

    if (cmd == "STOP") {
      bcode.Code = BC_STOP;
    }
    else if (regex_if2.MatchAt(cmd)) {
      if (regex_if2.GetMatchLength() != ~cmd)
        throw xRAMCompilerException(ECRCC_INCORRECT_SYNTAX, line);
      bcode.Code = BC_IF_NOT_NULL;
      bcode.Param1 = regex_if2.GetBackref(0);
      bcode.Param2 = regex_if2.GetBackref(1);
    }
    else if (regex_if1.MatchAt(cmd)) {
      if (regex_if1.GetMatchLength() != ~cmd)
        throw xRAMCompilerException(ECRCC_INCORRECT_SYNTAX, line);
      bcode.Code = BC_IF_NULL;
      bcode.Param1 = regex_if1.GetBackref(0);
      bcode.Param2 = regex_if1.GetBackref(1);
    }
    else if (regex_goto.MatchAt(cmd)) {
      if (regex_goto.GetMatchLength() != ~cmd)
        throw xRAMCompilerException(ECRCC_INCORRECT_SYNTAX, line);
      bcode.Code = BC_GOTO;
      bcode.Param1 = regex_goto.GetBackref(0);
    }
    else if (regex_sub.MatchAt(cmd)) {
      if (regex_sub.GetMatchLength() != ~cmd)
        throw xRAMCompilerException(ECRCC_INCORRECT_SYNTAX, line);
      bcode.Code = BC_SUB;
      bcode.Param1 = regex_sub.GetBackref(0);
      bcode.Param2 = regex_sub.GetBackref(1);
      bcode.Param3 = regex_sub.GetBackref(2);
    }
    else if (regex_add.MatchAt(cmd)) {
      if (regex_add.GetMatchLength() != ~cmd)
        throw xRAMCompilerException(ECRCC_INCORRECT_SYNTAX, line);
      bcode.Code = BC_ADD;
      bcode.Param1 = regex_add.GetBackref(0);
      bcode.Param2 = regex_add.GetBackref(1);
      bcode.Param3 = regex_add.GetBackref(2);
    }
    else if (regex_mov4.MatchAt(cmd)) {
      if (regex_mov4.GetMatchLength() != ~cmd)
        throw xRAMCompilerException(ECRCC_INCORRECT_SYNTAX, line);
      bcode.Code = BC_MOV_R_K;
      bcode.Param1 = regex_mov4.GetBackref(0);
      bcode.Param2 = regex_mov4.GetBackref(1);
    }
    else if (regex_mov3.MatchAt(cmd)) {
      if (regex_mov3.GetMatchLength() != ~cmd)
        throw xRAMCompilerException(ECRCC_INCORRECT_SYNTAX, line);
      bcode.Code = BC_MOV_RR_R;
      bcode.Param1 = regex_mov3.GetBackref(0);
      bcode.Param2 = regex_mov3.GetBackref(1);
    }
    else if (regex_mov2.MatchAt(cmd)) {
      if (regex_mov2.GetMatchLength() != ~cmd)
        throw xRAMCompilerException(ECRCC_INCORRECT_SYNTAX, line);
      bcode.Code = BC_MOV_R_RR;
      bcode.Param1 = regex_mov2.GetBackref(0);
      bcode.Param2 = regex_mov2.GetBackref(1);
    }
    else if (regex_mov1.MatchAt(cmd)) {
      if (regex_mov1.GetMatchLength() != ~cmd)
        throw xRAMCompilerException(ECRCC_INCORRECT_SYNTAX, line);
      bcode.Code = BC_MOV_R_R;
      bcode.Param1 = regex_mov1.GetBackref(0);
      bcode.Param2 = regex_mov1.GetBackref(1);;
    }
    else throw xRAMCompilerException(ECRCC_INVALID_CODE, line + 1);

    destination += bcode;
  }
}
