/*
  Extended string object
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : Extended string object
// ----------------------------------------------------------------------------
//  Remarks          :
//    iXiONmedia will no longer support standard C null-terminated strings,
//    except for special cases where using char* style strings has
//    _essential_ performance/memory advantages.
//    The xString object is intolerant in terms of out-of-bounds arguments.
//    You're likely to get an exception for everything you do wrong.
//
// ----------------------------------------------------------------------------
//  Modifications    :
//    Hardy     08.04.1999      bug in SetLength(). Buffer must not be null.
// ----------------------------------------------------------------------------
//  (c) Copyright 1998 by iXiONmedia, all rights reserved.
// ----------------------------------------------------------------------------




#ifndef XSTRING_H
#define XSTRING_H




#include "base.h"
#include "auto.h"
#include "numeric.h"




// Macros ---------------------------------------------------------------------
#define XSTR_NUMCHARS "0123456789ABCDEF"




// Case Conversion ------------------------------------------------------------
char chUpper(char c);
char chLower(char c);




// xString --------------------------------------------------------------------
class xString {
  protected:
  TSize                         Length;
  xArrayAllocator<char>         Buffer;
  static TBool                  FlagCaseSens;

  public:
  xString()
    : Length(0)
    { SetLength(0); }
  xString(xString const &source)
    : Length(0)
    { operator=(source); }
  xString(char const *source)
    : Length(0)
    { operator=(source); }
  xString(char const source)
    : Length(0)
    { operator=(source); }

  // Information, mode switching, backward compatibility
  TSize operator~() const
    { return Length; }
  static TBool CaseSensitive()
    { return FlagCaseSens; }
  static void CaseSensitive(TBool casesens)
    { FlagCaseSens = casesens; }
  operator char const *() const
    { return !Buffer; }

  // Assignment and concatenation
  xString &operator=(xString const &source);
  xString &operator=(char const *source);
  xString &operator=(char const source);

  xString &operator+=(xString const &source);
  xString &operator+=(char const *source);
  xString &operator+=(char const source);

  xString operator+(xString const &source) const
    { return xString(*this) += source; }
  xString operator+(char const *source) const
    { return xString(*this) += source; }
  xString operator+(char const source) const
    { return xString(*this) += source; }

  // Construction
  xString &Clear()
    {
      SetLength(0);
      return *this;
    }
  xString &MakeThis(TSize repeat,xString const &source);
  xString &MakeThis(TSize repeat,char const source);
  static xString Make(TSize repeat,xString const &source)
    { xString temp; temp.MakeThis(repeat,source); return temp; }
  static xString Make(TSize repeat,char const source)
    { xString temp; temp.MakeThis(repeat,source); return temp; }
  static xString PercentBar(TSize count,TSize total,TSize length = 40,
    char const full = '=',char const empty = ' ');

  // Comparison
  TBool operator==(xString const &operand) const;
  TBool operator!=(xString const &operand) const
    { return !operator==(operand); }
  TBool operator<(xString const &operand) const;
  TBool operator<=(xString const &operand) const
    { return operator==(operand) || operator<(operand); }
  TBool operator>=(xString const &operand) const
    { return operator==(operand) || !operator<(operand); }
  TBool operator>(xString const &operand) const
    { return !operator<=(operand); }

  TBool operator==(char const *operand) const
    { return operator==(xString(operand)); }
  TBool operator!=(char const *operand) const
    { return operator!=(xString(operand)); }
  TBool operator<(char const *operand) const
    { return operator<(xString(operand)); }
  TBool operator<=(char const *operand) const
    { return operator<=(xString(operand)); }
  TBool operator>=(char const *operand) const
    { return operator>=(xString(operand)); }
  TBool operator>(char const *operand) const
    { return operator>(xString(operand)); }

  // Extraction
  char &operator[](TIndex position) const;
  xString operator()(TDelta start) const;
  xString operator()(TDelta start,TSize length) const;
  xString &operator<<=(TSize maxlength)
    { return ClipThis(maxlength); }
  xString operator<<(TSize maxlength) const
    { return Clip(maxlength); }
  xString &operator>>=(TSize maxlength)
    {
      maxlength = NUM_MIN(maxlength,Length);
      return operator=(operator()(-maxlength));
    }
  xString operator>>(TSize maxlength) const
    {
      maxlength = NUM_MIN(maxlength,Length);
      return operator()(-maxlength);
    }

  // Insertion and Deletion
  xString Insert(TIndex position,xString const &substr) const;
  xString &InsertThis(TIndex position,xString const &substr)
    { return operator=(Insert(position,substr)); }
  xString Overwrite(TIndex position,xString const &substr) const;
  xString &OverwriteThis(TIndex position,xString const &substr)
    { return operator=(Overwrite(position,substr)); }
  xString Delete(TIndex start,TSize size) const;
  xString &DeleteThis(TIndex start,TSize size)
    { return operator=(Delete(start,size)); }
  xString Delete(TSize size) const
    { return Delete(0,size); }
  xString &DeleteThis(TSize size)
    { return DeleteThis(0,size); }
  xString DeleteChar(char ch) const;
  xString &DeleteCharThis(char ch)
    { return operator=(DeleteChar(ch)); }
  xString DeleteRight(TSize size = 1) const;
  xString &DeleteRightThis(TSize size = 1)
    { return operator=(DeleteRight(size)); }

  // Tools
  xString &ClipThis(TSize maxlength);
  xString Clip(TSize maxlength) const;
  xString Limit(TSize length,char filler = ' ') const;
  xString &LimitThis(TSize length,char filler = ' ')
    { return operator=(Limit(length,filler)); }
  xString PreLimit(TSize length,char filler = ' ') const;
  xString &PreLimitThis(TSize length,char filler = ' ')
    { return operator=(PreLimit(length,filler)); }
  xString Reverse() const;
  xString &ReverseThis()
    { return operator=(Reverse()); }
  xString Optimize() const;
  xString &OptimizeThis()
    { return operator=(Optimize()); }
  TBool InSet(char const &ch) const
    { return CountChar(ch) > 0; }
  TBool InSet(xString const &set) const;

  // Finding
  TSize CountChar(char const ch) const;
  TIndex Locate(xString const &searchstr,TIndex startat = 0) const;
  TIndex Locate(char const searchchar,TIndex startat = 0) const;
  TIndex LocateLast(xString const &searchstr) const;
  TIndex LocateLast(char const searchchar) const;
  xString Replace(xString const &source,xString const &destination) const;
  xString &ReplaceThis(xString const &source,xString const &destination)
    { return operator=(Replace(source,destination)); }

  // Case conversion
  xString &LowerThis();
  xString &UpperThis();
  xString Lower() const
    { xString temp = *this; return temp.LowerThis(); }
  xString Upper() const
    { xString temp = *this; return temp.UpperThis(); }

  // Numeric conversion
  static xString Unsigned2Base(unsigned long value,char digits = 0,char radix = 10);
  static xString Unsigned2Dec(unsigned long value,char digits = 0)
    { return Unsigned2Base(value,digits,10); }
  static xString Unsigned2Hex(unsigned long value,char digits = 0)
    { return Unsigned2Base(value,digits,16); }
  static xString Unsigned2Bin(unsigned long value,char digits = 0)
    { return Unsigned2Base(value,digits,2); }
  static xString Unsigned2Oct(unsigned long value,char digits = 0)
    { return Unsigned2Base(value,digits,8); }
  static xString Signed2Base(signed long value,char digits = 0,char radix = 10);
  static xString Signed2Dec(signed long value,char digits = 0)
    { return Signed2Base(value,digits,10); }
  static xString Signed2Hex(signed long value,char digits = 0)
    { return Signed2Base(value,digits,16); }
  static xString Signed2Bin(signed long value,char digits = 0)
    { return Signed2Base(value,digits,2); }
  static xString Signed2Oct(signed long value,char digits = 0)
    { return Signed2Base(value,digits,8); }

  xString &Unsigned2BaseThis(unsigned long value,char digits = 0,char radix = 10)
    { return *this = Unsigned2Base(value,digits,radix); }
  xString &Unsigned2DecThis(unsigned long value,char digits = 0)
    { return *this = Unsigned2Base(value,digits,10); }
  xString &Unsigned2HexThis(unsigned long value,char digits = 0)
    { return *this = Unsigned2Base(value,digits,16); }
  xString &Unsigned2BinThis(unsigned long value,char digits = 0)
    { return *this = Unsigned2Base(value,digits,2); }
  xString &Unsigned2OctThis(unsigned long value,char digits = 0)
    { return *this = Unsigned2Base(value,digits,8); }
  xString &Signed2BaseThis(signed long value,char digits = 0,char radix = 10)
    { return *this = Signed2Base(value,digits,radix); }
  xString &Signed2DecThis(signed long value,char digits = 0)
    { return *this = Signed2Base(value,digits,10); }
  xString &Signed2HexThis(signed long value,char digits = 0)
    { return *this = Signed2Base(value,digits,16); }
  xString &Signed2BinThis(signed long value,char digits = 0)
    { return *this = Signed2Base(value,digits,2); }
  xString &Signed2OctThis(signed long value,char digits = 0)
    { return *this = Signed2Base(value,digits,8); }

  static xString Bytes2Dec(TSizeD bytes);
  unsigned long EvalUnsigned() const;
  signed long EvalSigned() const;

  protected:
  void SetLength(TSize length = 0)
    {
      if (length+1 > ~Buffer) Buffer(length+length / 3+1);
      Buffer[length] = 0;
      Length = length;
    }
  unsigned long Evaluate(xString const &numstr,char radix = 10) const;

  friend xString operator+(char const *s1,xString const &s2);
  friend xString operator+(char const s1,xString const &s2);
};




#endif
