/*
  Regular expressions string object.
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : Regular expressions string object.
// ----------------------------------------------------------------------------
//  Remarks          :
//    xRegexMatcher::Match must set the MatchLength member correctly.
//    Hopefully, one day this will have full perl5 functionality.
//    Currently, you cannot quantify anything that generates backreferences,
//    not even alternatives.
//    Backref numbering starts at \0.
//    If you query anything about the last match, and that last match did
//    never happen, behavior is undefined.
//    xRegexString::ReplaceAll does not set the MatchIndex/MatchGlobal
//    members.
//
// ----------------------------------------------------------------------------
//  (c) Copyright 1998 by iXiONmedia, all rights reserved.
// ----------------------------------------------------------------------------




#ifndef XSTRRE_H
#define XSTRRE_H




#include "xstring.h"
#include "ctlist.h"




// Regex exceptions -----------------------------------------------------------
#define ECRE_INVQUANTIFIER              0
#define ECRE_UNBALBACKREF               1
#define ECRE_INVESCAPE                  2
#define ECRE_INVBACKREF                 3
#define ECRE_UNTERMCLASS                4




class xRegexException : public xBaseException {
  public:
  xRegexException(TErrorCode error,
    char const *info = NULL,char *module = NULL,TIndexD line = 0);
  virtual char *GetText() const;
};




// Regex token macros ---------------------------------------------------------
#define XSTRRE_LITERAL          '\\'
#define XSTRRE_BACKREF          '\\'
#define XSTRRE_ESCAPESEQ        '\\'
#define XSTRRE_ANYCHAR          '.'
#define XSTRRE_START            '^'
#define XSTRRE_END              '$'
#define XSTRRE_ALTERNATIVE      '|'
#define XSTRRE_CLASSSTART       '['
#define XSTRRE_CLASSNEG         '^'
#define XSTRRE_CLASSRANGE       '-'
#define XSTRRE_CLASSSTOP        ']'

#define XSTRRE_BACKREFSTART     '('
#define XSTRRE_BACKREFSTOP      ')'

#define XSTRREQ_0PLUS           '*'
#define XSTRREQ_1PLUS           '+'
#define XSTRREQ_01              '?'
#define XSTRREQ_START           '{'
#define XSTRREQ_RANGE           ','
#define XSTRREQ_STOP            '}'
#define XSTRREQ_NONGREEDY       '?'




// Backref helper classes -----------------------------------------------------
class xRegexBackref {
  TBool         Ready;
  TIndex        At;
  TSize         Size;

  public:
  void Checkpoint(TIndex at)
    {
      At = at;
      Ready = FALSE;
    }
  void MakeReady(TIndex end)
    {
      Size = end-At;
      Ready = TRUE;
    }
  xString Get(xString const &candidate) const
    {
      if (!Ready) EX_THROW(Regex,ECRE_UNBALBACKREF)
      return candidate(At,Size);
    }
};




typedef xList<xRegexBackref> xRegexBackrefHeap;
typedef xListIterator<xRegexBackref> xRegexBackrefHeapIterator;




// Matcher classes ------------------------------------------------------------
class xRegexMatcher {
  protected:
  xRegexMatcher         *Next;
  TSize                 MatchLength;
  TBool                 DeleteNext;

  public:
  xRegexMatcher()
    : Next(NULL)
    { }
  virtual ~xRegexMatcher()
    { if (Next && DeleteNext) delete Next; }
  TSize GetMatchLength() const
    { return MatchLength; }
  TSize CollectMatchLength() const;
  xRegexMatcher *GetNext() const
    { return Next; }
  void SetNext(xRegexMatcher *next,TBool deletenext = TRUE)
    {
      Next = next;
      DeleteNext = deletenext;
    }
  virtual TBool Match(xRegexBackrefHeap &brheap,
    xString const &candidate,TIndex at)
    = 0;

  protected:
  TBool MatchNext(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at) const
    { return Next ? Next->Match(brheap,candidate,at) : YES; }
};




class xRegexQuantifier : public xRegexMatcher {
  TBool                 Greedy,MaxValid;
  TSize                 MinCount,MaxCount;
  xRegexMatcher         *Quantified;

  public:
  xRegexQuantifier(TBool greedy,TSize mincount)
    : Greedy(greedy),MaxValid(NO),MinCount(mincount)
    { }
  xRegexQuantifier(TBool greedy,TSize mincount,TSize maxcount)
    : Greedy(greedy),MaxValid(YES),MinCount(mincount),MaxCount(maxcount)
    { }
  virtual ~xRegexQuantifier()
    { if (Quantified) delete Quantified; }

  void SetQuantified(xRegexMatcher *quantified)
    { Quantified = quantified; }
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at);

  protected:
  TBool MatchQuantified(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at)
    { return Quantified ? Quantified->Match(brheap,candidate,at) : YES; }
  TSize GetQuantifiedMatchLength()
    { return Quantified ? Quantified->CollectMatchLength() : 0; }
  TBool InRange(TSize matchcount)
    {
      return MaxValid ?
        (matchcount >= MinCount) && (matchcount <= MaxCount) :
        (matchcount >= MinCount);
    }
  TBool MatchCount(xRegexBackrefHeap &brheap,xString const &candidate,TIndex at,
    TSize matchcount);
};




class xRegexStringMatcher : public xRegexMatcher {
  xString               MatchStr;

  public:
  xRegexStringMatcher(xString const &matchstr)
    : MatchStr(matchstr)
    { MatchLength = ~MatchStr; }
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at)
    {
      if (at+~MatchStr > ~candidate) return NO;
      return (candidate(at,~MatchStr) == MatchStr) &&
        MatchNext(brheap,candidate,at+~MatchStr);
    }
};




class xRegexAnyCharMatcher : public xRegexMatcher {
  public:
  xRegexAnyCharMatcher()
    { MatchLength = 1; }
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at)
    { return MatchNext(brheap,candidate,at+1); }
};




class xRegexStartMatcher : public xRegexMatcher {
  public:
  xRegexStartMatcher()
    { MatchLength = 0; }
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at)
    { return (at == 0) && MatchNext(brheap,candidate,at); }
};




class xRegexEndMatcher : public xRegexMatcher {
  public:
  xRegexEndMatcher()
    { MatchLength = 0; }
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at)
    { return (at == ~candidate) && MatchNext(brheap,candidate,at); }
};




class xRegexClassMatcher : public xRegexMatcher {
  xString CharClass;
  TBool Negated;

  public:
  xRegexClassMatcher(xString const &charclass)
    : CharClass(charclass),Negated(NO)
    {
      MatchLength = 1;
      if (~CharClass && CharClass[0] == XSTRRE_CLASSNEG) {
        CharClass = CharClass(1);
        Negated = YES;
      }
      ExpandClass();
    }
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at)
    {
      try {
        if (at >= ~candidate) return NO;
        return
          (Negated ? !CharInClass(candidate[at]) : CharInClass(candidate[at])) &&
          MatchNext(brheap,candidate,at+1);
      }
      EX_CATCHCODE(Generic,EC_INDEX,return NO;)
    }

  private:
  TBool CharInClass(char const ch) const
    { return CharClass.InSet(ch); }
  void ExpandClass();
};




class xRegexBackrefOpenMatcher : public xRegexMatcher {
  public:
  xRegexBackrefOpenMatcher()
    { MatchLength = 0; }
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at);
};




class xRegexBackrefCloseMatcher : public xRegexMatcher {
  public:
  xRegexBackrefCloseMatcher()
    { MatchLength = 0; }
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at);
};




class xRegexAlternativeMatcher : public xRegexMatcher {
  class xRegexConnector : public xRegexMatcher {
    xRegexMatcher **pNext;

    public:
    xRegexConnector(xRegexMatcher **pnext)
      : pNext(pnext)
      { MatchLength = 0; }
    virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
      TIndex at)
      {
        Next = *pNext;
        return MatchNext(brheap,candidate,at);
      }
  };

  xList<xRegexMatcher *> AltList;
  xRegexConnector        Connector;

  public:
  xRegexAlternativeMatcher()
    : Connector(&Next)
    { }
  ~xRegexAlternativeMatcher();
  void operator+=(xRegexMatcher *matcher);
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at);
};




class xRegexBackrefMatcher : public xRegexMatcher {
  TIndex Backref;

  public:
  xRegexBackrefMatcher(TIndex backref)
    : Backref(backref)
    { }
  virtual TBool Match(xRegexBackrefHeap &brheap,xString const &candidate,
    TIndex at);
};




// The Regex class ------------------------------------------------------------
class xRegexString : public xString {
  typedef xString       Inherited;
  xRegexMatcher         *ParsedRegex;
  xRegexBackrefHeap     BackrefHeap;
  xString               LastCandidate;
  TIndex                MatchIndex;
  TSize                 MatchLength;

  public:
  xRegexString()
    : ParsedRegex(NULL),MatchIndex(0),MatchLength(0)
    {}
  xRegexString(xString const &source)
    : Inherited(source),ParsedRegex(NULL),MatchIndex(0),MatchLength(0)
    { }
  xRegexString(char const *source)
    : Inherited(source),ParsedRegex(NULL),MatchIndex(0),MatchLength(0)
    { }
  xRegexString(char const source)
    : Inherited(source),ParsedRegex(NULL),MatchIndex(0),MatchLength(0)
    { }
  ~xRegexString()
    { if (ParsedRegex) delete ParsedRegex; }

  xRegexString &Parse()
    {
      ParsedRegex = ParseRegex(*this);
      return *this;
    }
  TBool Match(xString const &candidate,TIndex from = 0);
  TBool MatchAt(xString const &candidate,TIndex at = 0);
  xString ReplaceAll(xString const &candidate,xString const &replacement,
    TIndex from = 0);

  // Queries pertaining to the last match
  TIndex GetMatchIndex()
    { return MatchIndex; }
  TSize GetMatchLength()
    { return MatchLength; }
  xString GetMatch()
    { return LastCandidate(MatchIndex,MatchLength); }
  xString GetBackref(TIndex index)
    { return BackrefHeap[index].Get(LastCandidate); }

  protected:
  xRegexMatcher *ParseRegex(xString const &expr);
  xRegexQuantifier *ParseQuantifier(xString const &expr,TIndex &at);
  TBool IsGreedy(xString const &expr,TIndex &at);
};




#endif

